#Copyright Angel Yanguas-Gil, 2014

"""Python Fresnel package

Fresnel is a python package that calculates coherent transmission
and reflexion of light from a multilayer structure.

Modules_

    base: contains the methods required to determine the reflection and
    transmission coefficients of interfaces and stacks
    optutils: conversion functions
    ellip: methods to model the spectroscopic ellipsometry response of
    a multilayer system

Methods_

    fresnel
    evtonm
    nmtoev
    ntoep
    epton

"""

from fresnel.base import fresnel, fresnel2, rs, rp
from fresnel.optutils import evtonm, nmtoev, ntoep, epton




# Copyright Angel Yanguas-Gil, 2014
"""
optutils: conversion functions

"""

import math as m
import cmath as cm

def radtodeg(a):
    """Radian to degree conversion

    """
    return 180*a/m.pi

def degtorad(a):
    """Degree to radian conversion

    """
    return m.pi*a/180.

def makeindex(n, k):
    """Return the complex refraction index from the real index and
    the attenuation constant
    
    """
    return n + 1J*k

def epton(ep):
    """dielectric constant to (complex) refraction index conversion
    
    """ 
    e1 = ep.real
    e2 = ep.imag
    t2 = m.sqrt(e1*e1+e2*e2)
    n1 = m.sqrt(0.5*(e1+t2))
    n2 = m.sqrt(0.5*(t2-e1))
    return n1+1J*n2

def ntoep(n):
    """(complex) refraction index to dielectric constant conversion
    
    """
    return n*n

def evtonm(en):
    """eV to nm conversion
    
    """
    return 1240./en

def nmtoev(nm):
    """nm to eV conversion
    
    """
    return 1240./nm

def hztoev(hz):
    """Frequency in Hz to energy in eV conversion
    
    """
    return hz*4.136e-15

def evtohz(en):
    """Energy in eV to frequency in Hz conversion
    
    """
    return en/4.136e-15

def cmtoev(k):
    """Wavenumber in cm-1 to energy in eV conversion
    
    """
    return 1.240e-4*k

def evtocm(en):
    """Eenergy in eV to wavenumber in cm-1 conversion
    
    """
    return 8065*en




#Copyright Angel Yanguas-Gil, 2014
"""
base: Fresnel conefficients and reflection and transmission
of polarized light from interfaces and multilayers determined using
the transfer matrix method.

"""
from __future__ import division

import cmath as cm
import math as m

try:
    from optutils import makeindex, radtodeg, degtorad
except ImportError:
    def makeindex(n0, n1):
        return n0 + 1J*n1

    def radtodeg(a):
        return m.pi*a/180.

    def degtorad(a):
        return a*180./m.pi


def fresnel(t0, wl, layers, deg=False):
    """
    Determine the fresnel coefficients of a multilayer.

    Arguments:
    t0: incident angle (in radians or degrees if deg=True)
    wl: wavelength (same units as thickness)
    layers: list [(n, k), (n, k, thickness),..., (n, k)]
    deg: flag indicating whether the incidence angle is in degrees.
    
    Return the tuple (rp, rs, tp, ts) with the reflection
    and transmission Fresnel coefficients for the p and s
    polarizations.

    """

    if deg == True:
        t0 = t0*m.pi/180.

#    if wl <= 0 or t0 >= 0.5*m.pi or t0 <= -0.5*m.pi:
#        raise ValueError("Parameter out of range")

    if len(layers) == 1:
        return (0.+0J, 0.+0J, 1.+0J, 1.+0J)
    elif len(layers) == 2:
        return fresnel2(t0, makeindex(layers[0][0], layers[0][1]),
            makeindex(layers[1][0], layers[1][1]))
    else:
        return _fresneln(t0, wl, layers)


def rp(t0, N0, N1, deg=False):
    """
    Calculate the reflection coefficient for p polarization
    
    Arguments:
    t0: incident angle (in radians or degrees if deg=True)
    N0: complex refraction index of the incident medium (n+ik)
    N1: complex refraction index of the second medium (n+ik)
    deg: flag controlling the incidence angle units


    Return rp, rs, tp, ts for the p and s polarizations.
    
    """

    if deg == True:
        t0 = t0*m.pi/180.
    ct0 = cm.cos(t0)
    st0 = cm.sin(t0)
    st1 = N0/N1*st0
    ct1 = cm.sqrt(1-st1*st1)
    rp = (N1*ct0 - N0*ct1)/(N1*ct0 + N0*ct1)
    return rp
    
def rs(t0, N0, N1, deg=False):
    """
    Calculate the reflection coefficient for p polarization
    
    Arguments:
    t0: incident angle (in radians or degrees if deg=True)
    N0: complex refraction index of the incident medium (n+ik)
    N1: complex refraction index of the second medium (n+ik)
    deg: flag controlling the incidence angle units


    Return rp, rs, tp, ts for the p and s polarizations.
    
    """

    if deg == True:
        t0 = t0*m.pi/180.
    ct0 = cm.cos(t0)
    st0 = cm.sin(t0)
    st1 = N0/N1*st0
    ct1 = cm.sqrt(1-st1*st1)
    rs = (N0*ct0 - N1*ct1)/(N0*ct0 + N1*ct1)
    return rs

def fresnel2(t0, N0, N1, deg=False):
    """
    Calculate the Fresnel coefficients for an interface.
    
    Arguments:
    t0: incident angle (in radians or degrees if deg=True)
    N0: complex refraction index of the incident medium (n+ik)
    N1: complex refraction index of the second medium (n+ik)
    deg: flag controlling the incidence angle units


    Return rp, rs, tp, ts for the p and s polarizations.
    
    """

    if deg == True:
        t0 = t0*m.pi/180.
    ct0 = cm.cos(t0)
    st0 = cm.sin(t0)
    st1 = N0/N1*st0
    ct1 = cm.sqrt(1-st1*st1)
    rp = (N1*ct0 - N0*ct1)/(N1*ct0 + N0*ct1)
    rs = (N0*ct0 - N1*ct1)/(N0*ct0 + N1*ct1)
    tp = (2.*N0*ct0)/(N1*ct0 + N0*ct1)
    ts = (2.*N0*ct0)/(N0*ct0 + N1*ct1)
    return rp, rs, tp, ts


def _fresneln(t0, wl, layers):
    """
    
    Determine the transmission and reflection coefficients of 
    fresnel coefficients of a multilayer for two or
    more interfaces.

    Arguments:
    t0: incident angle (in radians or degrees if deg=True)
    wl: wavelength (same units as thickness)
    layers: list [(n, k), (n, k, thickness),..., (n, k)]
    
    It returns the tuple (rp, rs, tp, ts) with the reflection
    and transmission Fresnel coefficients for the p and s
    polarizations.
    
    """

    N0 = makeindex(layers[0][0], layers[0][1])
    NS = makeindex(layers[-1][0], layers[-1][1])
    layers = layers[1:-1]
    
    mlp = []
    mls = []
    Nold = N0
    told = t0
    for layer in layers:
        Nnew = makeindex(layer[0], layer[1])
        mp, ms = _matinterface(Nold, Nnew, told)
        tnew = cm.asin(Nold*cm.sin(told)/Nnew)
        dnorm = layer[2]/wl*cm.cos(tnew)
        ml = _matlayer(Nnew, dnorm)
        mlp.append(mp)
        mls.append(ms)
        mlp.append(ml)
        mls.append(ml)
        Nold = Nnew
        told = tnew

    mp, ms = _matinterface(Nold, NS, told)
    mlp.append(mp)
    mls.append(ms)

    mp = _nmatmult(mlp)
    ms = _nmatmult(mls)

    rp = mp[2]/mp[0]
    rs = ms[2]/ms[0]
    tp = 1./mp[0]
    ts = 1./ms[0]

    return rp, rs, tp, ts


def fresnelreverse(rp, rs, tp, ts):
    """
    Calculate the inverse fresnel coefficients for an interface
    given the fresnel coefficients in one direction

    Arguments:
    rp, rs, tp, ts: reflection and transmission fresnel coefficients
    for the p and s polarizations.

    Return the inverse Fresnel coefficients rp2, rs2, tp2, ts2
    
    """
    
    rp2 =- rp
    rs2 =- rs
    tp2 = (1.-rp*rp)/tp
    ts2 = (1.-rs*rs)/ts
    return rp2, rs2, tp2, ts2


def _matinterface(N0, Nl, t0):
    """Return the matrix for an interface between media with complex 
    refractive indices N0 and N1 and incidence angle t0.
    
    """
    
    rp, rs, tp, ts = fresnel2(t0, N0, Nl)
    m1p = 1./tp
    m1s = 1./ts
    m2p = rp/tp
    m2s = rs/ts
    mp = (m1p, m2p, m2p, m1p)
    ms = (m1s, m2s, m2s, m1s)
    return mp, ms

def _matlayer(N, dn):
    """Return the matrix for a layer of refractive index N and 
    thickness (normalized to the vacuum wavelength) dn.
    
    """

    beta = 2.*m.pi*dn*N
    b1 = cm.exp(-1J*beta)
    b2 = cm.exp(1J*beta)
    return (b1, 0., 0., b2)

def _nmatmult(ml):
    """Multiply a list of 2x2 matrices coded as 4-tuples."""
    return reduce(_matmult, ml)

def _matmult(m1, m2):
    """Multiply two 2x2 matrices presented as 4-tuples."""
    m11 = m1[0]*m2[0]+m1[1]*m2[2]
    m12 = m1[0]*m2[1]+m1[1]*m2[3]
    m21 = m1[2]*m2[0]+m1[3]*m2[2]
    m22 = m1[2]*m2[1]+m1[3]*m2[3]
    return m11, m12, m21, m22

if __name__ == '__main__':
    mats = [(1., 0.), (2., 3., 100.), (1., 0.)]
    rp, rs, tp, ts = fresnel(0., 400., mats)
    print tp
    print ts




#Copyright Angel Yanguas-Gil, 2014
"""
ellip: Methods to simulate and manipulate spectroscopic ellipsometry data
"""

import math as m

from fresnel import fresnel
from fresnel.optutils import degtorad, radtodeg

def ellip(angle, waveln, stack, deg=True):
    """Return the ratio of the reflection Fresnel coefficients
    for a stack for a given wavelength and incident angle.

    The stack is formed by a list of tuples (n, k, thickness). In
    the case of the incident and transmitted media, thickness is
    ignored and (n, k) tuples can be passed instead.

    """
    
    rp, rs, tp, ts = fresnel(angle, waveln, stack, deg)
    return rp/rs


def anglestorho(psi, delta):
    """Calculate the ratio of the reflection Fresnel coefficients 
    from the ellipsometric angles psi and delta (in degrees)

    """

    psi = degtorad(psi)
    delta = degtorad(delta)
    tp = m.tan(psi)
    return complex(tp*m.cos(delta), tp*m.sin(delta))


def rhotoangles(rho):
    """Calculate the ellipsometric angles psi and delta from the ratio
    of the fresnel reflection coefficients rp/rs for p and s polarizations

    """

    r = abs(rho)
    psi = m.atan(r)
    psi = radtodeg(psi)
    delta = m.acos(rho.real/r)
    delta = radtodeg(delta)
    if rho.imag/r < 0:
        delta = 2.*m.pi - delta
    return psi, delta


def rhotoepsilon(inc, rho, deg=False):
    """Calculate the pseudodielectric constant given the angle of
    incidence (in radians unless deg flag is set to True)
    and rho, defined as the ratio of the reflection fresnel
    coefficients for p and s polarizations rp/rs

    """

    if deg:
        inc = degtorad(inc)

    try:
        ep = m.tan(inc)**2*(1.-4.*rho*m.sin(inc)**2/(1.+rho)**2)
    except ZeroDivisionError:
        ep = None
    return ep



from distutils.core import setup
setup(name = 'fresnel',
    description = "light interaction with interfaces and multilayers" ,
    version = '0.1.1',
    author = "Angel Yanguas-Gil",
    author_email = "angel.yanguas@gmail.com",
    url = "http://ayanguasgil.net/",
    download_url = (
        "http://ayanguasgil.net/stuff/fresnel-0.1.1.tar.gz"),
    packages = ['fresnel'],
    classifiers = ["Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Topic :: Scientific/Engineering",
        "Topic :: Utilities",
        "License :: OSI Approved :: BSD License" ]
)

